/*
 * See http://www.iki.fi/~jylitalo/apache/mod_auth_samba.html for
 * introduction to module and its directives.
 *
 * copyright (c) 1997, 1998 by Juha Ylitalo (juha.ylitalo@iki.fi)
 *
 * This module is derived from work done by the Apache Group and is
 * subject to the same terms and conditions as the Apache server.
 */

#include "mod_auth_samba.h"
#include <ndbm.h>
#include <errno.h>
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "grp.h"

/*
 * Data struct, where all module specific information is stored.
 */
typedef struct mod_samba_auth_config_struct {
  int samba_auth_enabled_flag;
  int samba_auth_authorative_flag;
  int samba_auth_num_of_domains;
  int samba_auth_num_of_domain_controllers[ 16 ];
  char* samba_auth_domain_controllers[ 16 ][ 16 ]; /* name of NT server */
  char* samba_auth_domain_names[ 16 ];
  char* samba_auth_group_dbm;         /* database name     */
  char* samba_auth_passwd_cache;
} mod_samba_auth_config_rec;

/* "Constructor" */
void *mod_samba_auth_create_dir_config (pool *p, char *d)
{
    mod_samba_auth_config_rec *sec = ap_pcalloc( p , 
						 sizeof( mod_samba_auth_config_rec ) );
    sec->samba_auth_enabled_flag = 1;
    sec->samba_auth_authorative_flag = 1;
    sec->samba_auth_num_of_domains = 0;
    sec->samba_auth_passwd_cache = strdup( "/tmp/.mod_auth_samba" );
    sec->samba_auth_group_dbm = NULL;
    return(sec);
}

/*
 * Setter methods 
 */
const char *set_samba_auth_enabled_flag (cmd_parms *cmd,
                              mod_samba_auth_config_rec *sec, int arg) {
    sec->samba_auth_authorative_flag=arg;
    return NULL;
}

const char *set_samba_auth_authorative_flag (cmd_parms *cmd,
                              mod_samba_auth_config_rec *sec, int arg) {
    sec->samba_auth_authorative_flag=arg;
    return NULL;
}

const char *set_samba_auth_domain_controller( cmd_parms* cmd,
					      mod_samba_auth_config_rec* sec, 
					      char* arg) {
  int i,j;
  int num;
  char *tok;


  tok = strtok( arg , ":" );
  if ( sec->samba_auth_num_of_domains < 15 ) {
    num = sec->samba_auth_num_of_domains;
    sec->samba_auth_domain_names[ num ] = strdup( tok );
    tok = strtok( NULL , "," );
    for ( i = 0 ; i < 16 && tok != NULL ; i++ ) {
      sec->samba_auth_domain_controllers[ num ][ i ] = strdup( tok );
      tok = strtok( NULL , "," );
    }
    sec->samba_auth_num_of_domain_controllers[ num ] = i;
    sec->samba_auth_num_of_domains = num + 1;
  }
  /* 
     fprintf( stderr , "SUMMARY:" ); 
  for ( i = 0 ; i < sec->samba_auth_num_of_domains ; i++ ) {
    fprintf( stderr , "\n(%d:%d) %s : ", i , sec->samba_auth_num_of_domain_controllers[ i ] , sec->samba_auth_domain_names[ i ] ); 
    for ( j = 0 ; j < sec->samba_auth_num_of_domain_controllers[ i ] ; j++ ) 
      fprintf( stderr , "%s" , sec->samba_auth_domain_controllers[ i ][ j ] );
  }
  fprintf( stderr , "\n" );
  */

  return NULL;
}

const char *set_samba_auth_group_dbm( cmd_parms* cmd,
					   mod_samba_auth_config_rec* sec, 
					   char* arg) {
  if ( sec->samba_auth_group_dbm != NULL ) {
    free( sec->samba_auth_group_dbm );
  }
  sec->samba_auth_group_dbm = strdup( arg );
  return NULL;
}

const char *set_samba_auth_passwd_cache( cmd_parms* cmd,
					 mod_samba_auth_config_rec* sec, 
					 char* arg) {
  if ( sec->samba_auth_passwd_cache != NULL ) {
    free( sec->samba_auth_passwd_cache );
  }
  sec->samba_auth_passwd_cache = strdup( arg );
  return NULL;
}

/*
 * Data structure for defining which lines in .htaccess and 
 * Directory are important for us. Look Apache API documentation
 * for descriptions.
 */
command_rec mod_samba_auth_cmds[] = {
  { "AuthSambaEnabled" , set_samba_auth_enabled_flag , 
    (void *)NULL, OR_AUTHCFG, FLAG, 
    "When 'off', module will DECLINE from all authentitcation requests." },
  { "AuthSambaAuthoritative", set_samba_auth_authorative_flag,
    (void *)NULL, OR_AUTHCFG, FLAG,
    "When 'on' the system access files are taken to be authorative and access control is not passed along to other access modules." },
  { "AuthSambaDomain", set_samba_auth_domain_controller,
    (void *)NULL, OR_AUTHCFG, ITERATE,
    "NT server, which local user database is used for password authentication." },
  { "AuthSambaPasswdCache" , set_samba_auth_passwd_cache, (void*)NULL, OR_AUTHCFG, TAKE1, 
    "Name for dbm that caches users passwords." } ,
  { "AuthSambaGroupDbm", set_samba_auth_group_dbm, (void*)NULL, OR_AUTHCFG, TAKE1,
    "Group, where user should belong." },
  { NULL }
};
  
module samba_auth_module;

int mod_samba_auth_check_passwd_against_domain_controller( request_rec* r,
							   char* user , 
							   char* passwd ,
							   char* server ,
							   char* domain ,
							   char* dbase ) {
  datum package = { user , ( strlen( user ) + 1 ) * sizeof( char ) };
  datum dat;
  DBM* fout;
  char err_str[ MAX_STRING_LEN ];
  char data_str[ MAX_STRING_LEN ];
  int error_code;
  int ret;

  ret = Valid_User( user , passwd , server , NULL , domain );

  switch ( ret ) {
  case NTV_SERVER_ERROR:
    /* fprintf( stderr , "PAM_SMB : Server (%s) Error\n" , server );*/
    return -2;
  case NTV_PROTOCOL_ERROR:
    /* fprintf( stderr , "PAM_SMB : Protocol Error on %s\n" , server );*/
    return -2;
  case NTV_LOGON_ERROR:
    /* fprintf( stderr , "PAM_SMB : Login failed (%s::%s@%s)\n" , 
       domain , user , server ); */
    return -2;
    break;
  case 0: /* Here we store the username/password pair into cache */
    /* fprintf( stderr , "PAM_SMB : Login accepted (%s::%s@%s)\n" , 
       domain , user , server ); */
    fout = dbm_open( dbase , O_RDWR | O_CREAT , 0664 );
    if ( dbm_error( fout ) != 0 ) {
      /* fprintf( stderr , "dbm_open(%s) failed on user ( %s ) \n" , dbase , user ); */
    } else {
      /* In SUNOS4, we don't have to specify the seed in sprintf. */
#ifndef SUNOS4
      sprintf( data_str , "%ld:69%s" , time( NULL ) , crypt( passwd , "69" ) );
#else
      sprintf( data_str , "%ld:%s" , time( NULL ) , crypt( passwd , "69" ) );
#endif
      dat.dptr = strdup( data_str );
      dat.dsize = ( strlen( data_str ) +1 ) * sizeof( char );
      error_code = dbm_store( fout , package , dat , DBM_REPLACE );
      /* if ( error_code < 0 ) {
	fprintf( stderr , "dbm_store(%s) failed on user ( %s ) (errno = %d)\n" , dbase , user , errno );
      } else {
	fprintf( stderr , "user updated into cache.\n" );
	} */ /* if */
    } /* if */
    dbm_close( fout );
    break;
  } /* switch */
  return ret;
}

/*
 * Check if user - password pair is valid.
 *
 * @param r      for error messages
 * @param user   username
 * @param passwd password
 * @param sec    information about domain controller among other
 * @return  0 - user passes
 *         -1 - no match
 */
int mod_samba_auth_check_passwd( request_rec* r,
				 char* user , 
				 char* passwd , 
				 mod_samba_auth_config_rec* sec )
{
  int i , j , ret;
  char err_str[ MAX_STRING_LEN ];
  DBM* f = NULL;
  datum package = { user , ( strlen( user ) + 1 ) * sizeof( char ) };
  datum results;
  char* dbase = sec->samba_auth_passwd_cache;

  f = dbm_open( dbase , O_RDONLY , 0664 );
  if ( f != NULL && dbm_error( f ) == 0 ) {
    results = dbm_fetch( f , package );
    dbm_close( f );

    if ( results.dptr != NULL ) {
      char result_str[ MAX_STRING_LEN ];
      char salt[ 3 ];
      long now = time( NULL );
      long was;
      char* oldpasswd;
      char* newpasswd;

      strncpy( result_str , results.dptr , results.dsize / sizeof( char ) );
      
      oldpasswd = strchr( result_str , ':' );
      oldpasswd[ 0 ] = '\0';
      salt[ 0 ] = oldpasswd[ 1 ];
      salt[ 1 ] = oldpasswd[ 2 ];
      salt[ 2 ] = '\0';
      oldpasswd = &oldpasswd[ 3 ];
      was = atol( result_str );
      if ( ( was + 8 * 3600 ) >= now ) {
	newpasswd = crypt( passwd , salt );
	if ( strcmp( oldpasswd , newpasswd ) == 0 ) {
	  return( 0 ); /* And we have a winner :) */
	}
      }
    }
    /*  } else {
	fprintf( stderr , "AuthSamba failed to open cache ( errno = %d )\n" , errno ); */
  }

  /* Setup connection */
  if ( sec->samba_auth_num_of_domains == 0 ) {
    /* fprintf( stderr , "AuthSambaDomain is not defined.\n" ); */
    return( -1 );
  } /* if */

  for ( i = 0 , ret = -1; 
	i < sec->samba_auth_num_of_domains 
	  && ret != 0 ; 
	i++ ) {
      char* domain = sec->samba_auth_domain_names[ i ];

      for ( j = 0 , ret = -2; 
	   j < sec->samba_auth_num_of_domain_controllers[ i ] && ret == -2 ; 
	   j++ ) {
	  char* server = sec->samba_auth_domain_controllers[ i ][ j ];
	  
	  ret = mod_samba_auth_check_passwd_against_domain_controller( r , 
								      user , 
								      passwd , 
								      server ,
								      domain ,
								      dbase );
      } /*for (j) */
  } /* for (i) */
  return( ( ret == 0 ) ? 0 : -1 );
} /* mod_samba_auth_check_passwd */

/*
 * group check
 *
 * @param r    only used for error message reporting
 * @param sec  has information about dbm file
 * @param name name of group
 * @param user tested user
 * @return  0 - user passes
 *         -1 - no match
 */
int mod_samba_auth_in_group(request_rec* r , 
			    mod_samba_auth_config_rec* sec , 
			    char* name , 
			    char* user )
{
  static DBM* f = NULL;
  static time_t expires = 0;
  char full_list[ MAX_STRING_LEN ];
  int len;
  char* buf;
  datum package = { user , ( strlen( user ) + 1 ) * sizeof( char )  };
  datum results;

  if ( sec->samba_auth_group_dbm == NULL ) {
    return -1;
  }
  
  if ( ( f != NULL ) && ( time(NULL) > expires ) ) {
    dbm_close( f );
    f = NULL;
  }
  if ( f == NULL ) {
    expires = time(NULL) + 3600;
    f = dbm_open( sec->samba_auth_group_dbm , O_RDONLY , 0664 );
    
    if ( dbm_error( f ) != 0 ) {
      return -1;
    }
  } 

  results = dbm_fetch( f , package );
  if ( results.dptr !=  NULL ) {
    len = results.dsize / sizeof( char );
    strncpy( full_list , results.dptr , len );
    full_list[ len ] = '\0';
  } else {
    full_list[ 0 ] = '\0';
  }
  if ( results.dptr == NULL ) {
    return -1;
  }

  buf = strstr( full_list , name );
  if ( buf != NULL && buf[ -1 ] == ' ' && buf[ strlen(name) ] == ' ' ) {
    /* fprintf( stderr , "User belongs to right circles :)\n" ); */
    return(0);
  }
  /* fprintf( stderr , "User didn't belong to right circles :)\n" );*/
  return( -1 );
}

/* These functions return 0 if client is OK, and proper error status
 * if not... either AUTH_REQUIRED, if we made a check, and it failed, or
 * SERVER_ERROR, if things are so totally confused that we couldn't
 * figure out how to tell if the client is authorized or not.
 *
 * If they return DECLINED, and all other modules also decline, that's
 * treated by the server core as a configuration error, logged and
 * reported as such.
 */

/* Determine user ID, and check if it really is that user, for HTTP
 * basic authentication...
 */

int mod_samba_auth_authenticate_basic_user ( request_rec* r ) {
  mod_samba_auth_config_rec* sec =
    (mod_samba_auth_config_rec *)ap_get_module_config (r->per_dir_config,
						    &samba_auth_module);
  conn_rec *c = r->connection;
  char *sent_pw;
  int res;

  if ( !sec->samba_auth_enabled_flag ) {
    return DECLINED ;
  }
  if ( ( res = ap_get_basic_auth_pw( r , &sent_pw ) ) )
    return res;
  
  /* If it non-authorative and we don't find the user just pass it on. */
  if ( mod_samba_auth_check_passwd( r , c->user , sent_pw , sec ) != 0 ) {
    if ( !sec->samba_auth_authorative_flag ) {
      return DECLINED;
    }
    ap_note_basic_auth_failure( r );
    return AUTH_REQUIRED;
  }
  return OK;
}

int mod_samba_auth_check_auth (request_rec *r) {
  mod_samba_auth_config_rec* sec =
    ( mod_samba_auth_config_rec* )ap_get_module_config( r->per_dir_config ,
						      &samba_auth_module );
  char errstr[ MAX_STRING_LEN ];
  char* user = r->connection->user;
  int m = r->method_number;
  array_header* reqs_arr = ap_requires( r );
  require_line* reqs = reqs_arr ? ( require_line* ) reqs_arr->elts : NULL;
  register int x;
  const char* t;
  char* w;
  
  if ( !reqs_arr ) {
    if ( !sec->samba_auth_enabled_flag ) {
      return DECLINED;
    }
    if ( sec->samba_auth_authorative_flag ) {
      ap_note_basic_auth_failure( r );
      return AUTH_REQUIRED;
    };
    return DECLINED;
  }
  
  for(x=0; (x < reqs_arr->nelts) ; x++) {
    
    if (! (reqs[x].method_mask & (1 << m))) continue;
    
    t = reqs[x].requirement;
    w = ap_getword(r->pool, &t, ' ');
    
    if (!strcmp(w,"user")) {
      while(t[0]) {
	w = ap_getword_conf( r->pool , &t );

	if ( strcmp( user , w ) == 0 ) {
	  return( OK );
	} /* if */
      } /* while */
    } /* if */
    
    if (!strcmp( w , "group")) {
      /* look up the membership for each of the group_dbm in the table */
      while( t[0] ) {
	w = ap_getword_conf( r->pool , &t );
	if ( mod_samba_auth_in_group( r , sec , w , user ) == 0 ) {
	  return( OK );
	} /* if */
      } /* while */
    } /* if */
    
    if(!strcmp(w,"valid-user")) {
      return( OK );
    }
  }

  if ( sec->samba_auth_authorative_flag ) {
    ap_note_basic_auth_failure( r );
    return AUTH_REQUIRED;
  }
  return DECLINED;
}
    
module samba_auth_module = {
  STANDARD_MODULE_STUFF,
  NULL,			/* initializer */
  mod_samba_auth_create_dir_config,	/* dir config creater */
  NULL,			/* dir merger --- default is to override */
  NULL,			/* server config */
  NULL,			/* merge server config */
  mod_samba_auth_cmds,		/* command table */
  NULL,			/* handlers */
  NULL,			/* filename translation */
  mod_samba_auth_authenticate_basic_user,	/* check_user_id */
  mod_samba_auth_check_auth,	/* check auth */
  NULL,			/* check access */
  NULL,			/* type_checker */
  NULL,			/* fixups */
  NULL,			/* logger */
  NULL				/* header parser */
};
